"use strict";

// General app-state manager
// - stores all files to disk (in specified "state files" dir, within app sandbox dir (not sdcard)).
// - uses utilFileMgr internally
// - caches in memory, enabling async-or-sync method options
// - unlike utilAppState (uses localstorage w/ MB limits) can store >10MB statefiles.
// - values converted to JSON for storage, so values limited accordingly.
// - (mostly from utilStateMgr from gogy)
// - Note: Android permission (when sdk>=23) not needed, since we read/write only to app sandbox folder.

function utilStateFilesMgr(stateDirName) {
	this.test = "";
	this._fileMgr = new utilFileMgr();
	this._saveDirpath = stateDirName;

	this._queuedFileWrite_queue = [];
	this._StateCache = {};
}

// Init
utilStateFilesMgr.prototype.initStateMgr = function(callback) {
	var self = this;
	
	if (window.cordova) {
		if (g_isIOS)
			this._fileMgr.cordovaFileRoot = cordova.file.documentsDirectory;
		else {
			//this._fileMgr.cordovaFileRoot = cordova.file.externalRootDirectory;
			this._fileMgr.cordovaFileRoot = cordova.file.dataDirectory; // ("/data/data/<app-id>/files")
		}
	}
	this._fileMgr.chromeStorageType = window.PERSISTENT;
	this._fileMgr.chromeRequestQuotaBytes = 100*1024*1024; // 100MB
	
	self._fileMgr.initFileSys(function(rootDir) {
		console.log("filesys init'ed");

		// Create root directory.  (Need to call before first recording is saved.)
		//self._fileMgr.createDir(self._appdirname, function() {
			self._fileMgr.createDir(self._saveDirpath, function() {
				callback();
			});
		//});
	});
}


// Pre-load files up-front, for fileReadSync() later.
// Enables synchronous calls downstream.
utilStateFilesMgr.prototype.preloadFiles = function(idStrAr, callback) {
	var self = this;
	var numComplete = 0;
	var doInitData = function(idStr, cb2) {
		self.loadDataAsync(idStr, function(value) {
			cb2();
		});
	}
	// Read/write each dataKey once, so later reads can be called synchronously.
	for (var i=0; i<idStrAr.length; i++) {
		doInitData(idStrAr[i], function() {
			numComplete += 1;
			if (numComplete == idStrAr.length)
				callback();
		});
	}
}


// Queued/serial (safer) file write
// (Multiple quick-succession calls to _fileMgr.createJpegFile() results in a corrupt file. "abcdefgNULNULNULNULNULNUL" -- the 1st (smaller) truncates as the 2nd (bigger) is still (or just finished) writing.)
utilStateFilesMgr.prototype._queuedFileWrite = function(idStr, dataObjVal, callback) {
	var self = this;
	var isProcessing = this._queuedFileWrite_queue.length ? true : false;
	var queue = this._queuedFileWrite_queue;

	// Add to queue, synchronous.
	queue.push({ idStr: idStr, dataObjVal: dataObjVal, cb2: callback });

	// If queue has something in it, then it's actively processing.  Add to queue only.
	if (!isProcessing) {
		var processUntilDone = function() {
			//var queue = self._queuedFileWrite_queue;
			if (!queue.length) { 
				return; // Finished processing queue. Do nothing, no callback.
			}
			// Process first item, until done.
			//console.log("QUEUE");
			//console.log(queue);
			var item = queue[0];
			self._fileMgr.createJpegFile(self._saveDirpath, item.idStr + ".json", JSON.stringify(item.dataObjVal), function() {
				queue.shift(); // remove the first element, which has now been processed.
				item.cb2();
				processUntilDone(); // continue calling until queue is empty.
			});
		}
		processUntilDone();
	}
}

utilStateFilesMgr.prototype.saveDataAsync = function(idStr, dataObjVal, callback) {
	// Save arbitrary JS obj here (saved as JSON to filesystem)
	var self = this;
	self._StateCache[idStr] = dataObjVal;
	self._queuedFileWrite(idStr, dataObjVal, callback);
}


utilStateFilesMgr.prototype.loadDataAsync = function(idStr, callback) {
	var self = this;
	
	// Load saved JSON data, return as JS obj
	var output;
	// Try to load from cache (ram).
	if (idStr in self._StateCache) {
		output = self._StateCache[idStr];
		callback(output);
		return;
	}
	else {
		this._fileMgr.getFile(this._saveDirpath + "/" + idStr + ".json", true, function(fileEntry) {
			self._fileMgr.readFileEntryAsText(fileEntry, function(fileText) {
				if (!fileText) // convert all falseys (e.g. "" when file DNE) into parsable JSON.
					fileText = false;
				try {
					output = JSON.parse(fileText);
				}
				catch (e) {
					output = false;
					alert("Read invalid JSON from file: " + idStr);
					console.log("unparsable JSON: '" + fileText + "'");
				}
				// Cache in ram.
				self._StateCache[idStr] = output;
				callback(output);
				return;
			});
		});
	}
}

// Synchronous data load function.
utilStateFilesMgr.prototype.loadDataSync = function(idStr) {
	// Ensure data has been loaded from disk (must be asynchronously) already, by complaining loudly if hasn't been loaded yet.
	if (!(idStr in this._StateCache)) {
		throw "loadDataSync() called on uninitialized dataIDStr - '" + idStr + "'";
	}
	return this._StateCache[idStr];
}

// Synchronous data load function.
utilStateFilesMgr.prototype.saveDataSync = function(idStr, dataObjVal) {
	this._StateCache[idStr] = dataObjVal;
	// Kick off asynchronous file save.
	this.saveDataAsync(idStr, dataObjVal, function() {
		// Do nothing.  returned success to user before async call finishes.
	});
	return;
}

// Synchronous data load function.
utilStateFilesMgr.prototype.deleteDataAsync = function(idStr, callback) {
	var self = this;
	delete this._StateCache[idStr];
	this._fileMgr.deleteFile(this._saveDirpath + "/" + idStr + ".json", function() {
		console.log("Deleted file - " + idStr + ".json");
		// Load non-existent files into memory, for downstream loadDataSync() calls that may be depending on an upstream preload()/save() call.  Not un-goofy.
		self.preloadFiles([idStr], callback);
		//callback();
	});
	return;
}



utilStateFilesMgr.prototype.getNamesByPrefixAsync = function(idStrPrefix, callback) {
	this._fileMgr.dirContents(this._saveDirpath, function(fileEntries) {
		//console.log("DIR CONTENTS:");
		//console.log(fileEntries);
		var output = [];
		for (var i=0; i<fileEntries.length; i++) {
			if (fileEntries[i].isFile && fileEntries[i].name.startsWith(idStrPrefix)) {
				var name1 = fileEntries[i].name;
				if (name1.substr(name1.length - 5) == ".json") // remove ".json".
					name1 = name1.substr(0, name1.length - 5);
				output.push(name1);
			}
		}
		//console.log(output);
		callback(output);
	});
}

utilStateFilesMgr.prototype.deleteByPrefixAsync = function(idStrPrefix, callback) {
	var self = this;
	var deleteUntilDone = function(cursor_i, cacheNames, cb2) {
		if (cursor_i > cacheNames.length - 1) {
			cb2();
			return;
		}
		var cacheName = cacheNames[cursor_i];
		//console.log("deleting " + cacheName);
		self.deleteDataAsync(cacheName, function(value) {
			deleteUntilDone(cursor_i+1, cacheNames, cb2);
		});
	}
	//console.log("PREFIX " + idStrPrefix);
	this.getNamesByPrefixAsync(idStrPrefix, function(cacheNames) {
		//console.log("NAMES " + JSON.stringify(cacheNames));
		deleteUntilDone(0, cacheNames, callback);
	});
	return;
}
